/*
 * garux.c
 *
 * Linux Loader for the Palm OS.
 *
 * GaruxNG is Copyright (C) 2006 SCL (scl@hexview.com)
 * GaruxNG is a complete rewrite of Garux to enable
 * loading of kernel images from PalmOS card volumes
 *
 * Garux is Copyright (C) 2004 Vinayak R. Borkar
 * Author: Vinayak R. Borkar [vinayakb@users.sourceforge.net]
 * Original version by Romain Goyet [r.goyet@gmail.com]
 *
 * 11-Sep-2005, Alex Osborne [bobofdoom@gmail.com]:
 *   Updated to work with multiple handhelds.
 *   Device support based on the work of:
 *     Nikolay Petukhov   - Palm Tungsten T2
 *     Vladimir Pouzanov  - Palm Tungsten T3
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <PalmOS.h>
#include <PceNativeCall.h>
#include <VFSMgr.h>
#include <MemoryMgr.h>
#include "garux_r.h"
#include "config.h"

int main_event_loop();
int runLinuxKernel();
void SetFieldTextFromStr( FieldPtr fldP, char *strP, Boolean redraw);
void SetupMainForm(FormPtr frmP);
void Log(char *str);
void OnPortListChange();

FieldPtr kernelParametersField;
UInt16 krnl_volRefNum=0;
UInt32 krnl_size=0;
UInt32 initrd_size=0;
/* --------------------------------------------------------------*/
void Log(char *str){
char buf[2048];
FormPtr frmP=FrmGetActiveForm();
FieldPtr fldP= FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, LogField));
if (FldGetTextPtr(fldP)) {
    StrCopy(buf,FldGetTextPtr(fldP));
    StrCat(buf,"\n");
    StrCat(buf,str);
    SetFieldTextFromStr(fldP, buf, true);
}else
    SetFieldTextFromStr(fldP, str, true);
}
/* --------------------------------------------------------------*/
int main_event_loop() {
  UInt16 error_code;
  EventType event;
  
  while (1) {
    EvtGetEvent(&event, 100);
    if (SysHandleEvent(&event))
      continue;
    if (MenuHandleEvent(0, &event, &error_code))
      continue;
	
    if (event.eType == ctlSelectEvent) {
	if (event.data.ctlSelect.controlID == LinuxButton) {
#ifndef	SUPPRESS_WARNING
    	    if (FrmAlert(StartupAlert) == 1) {
	    FrmCloseAllForms();
	    return 0;
    	    }
#endif
	runLinuxKernel();
	return 0;
        }
    }

    if (event.eType == popSelectEvent) {
	OnPortListChange();

        continue;
    }
    
    if (event.eType == frmLoadEvent) {
      FrmSetActiveForm(FrmInitForm(event.data.frmLoad.formID));
      continue;
    }
    if (event.eType == frmOpenEvent) {
      SetupMainForm(FrmGetActiveForm());
      continue;
    }
    if (event.eType == appStopEvent) {
      FrmCloseAllForms();
      return 0;	
    }
    if (FrmGetActiveForm())
      FrmHandleEvent(FrmGetActiveForm(), &event);
  }
  
}

/* --------------------------------------------------------------*/
UInt32 PilotMain (UInt16 cmd, void *params, UInt16 flags) {
  if (sysAppLaunchCmdNormalLaunch == cmd) {
    FrmGotoForm(MainForm);
    return main_event_loop();
    }
    return 0;
}
/* --------------------------------------------------------------*/
void SetupMainForm(FormPtr frmP){
  char buf[256];
  ControlPtr ctlP;
  ListPtr lstP;
  char vname[32];
  UInt16 volRefNum;
  UInt32 volIterator=vfsIteratorStart;
  FileRef fRef;
  UInt32 sz;
  Err err;
  
    ctlP=FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, LinuxButton));
    CtlSetEnabled(ctlP,false);
     FrmDrawForm(FrmGetActiveForm());

    while ((volIterator != vfsIteratorStop)) {
	if (VFSVolumeEnumerate(&volRefNum, &volIterator) == errNone) {
	    err=VFSFileOpen(volRefNum,"/zImage",vfsModeRead,&fRef);
	if (err==errNone) {
	    MemSet(vname,0,sizeof(vname));
    	    VFSVolumeGetLabel(volRefNum,vname,sizeof(vname)-1);
	    VFSFileSize(fRef,&sz);
            VFSFileClose(fRef);
	    StrPrintF(buf,"[%s] kernel: %ld bytes",vname,sz);Log(buf);
	    volIterator = vfsIteratorStop;
	    krnl_volRefNum=volRefNum;krnl_size=sz;
	    err=VFSFileOpen(volRefNum,"/initrd.gz",vfsModeRead,&fRef);
	    if (err==errNone) {
		VFSFileSize(fRef,&sz);
        	VFSFileClose(fRef);
		StrPrintF(buf,"[%s] initrd: %ld bytes",vname,sz);Log(buf);
		initrd_size=sz;
		}
	    else {
    		Log("initrd: not found");
		}
	    
	    }
	}
    }

    if ((krnl_size==0)||(initrd_size==0)) {
	Log("\n* Place zImage and initrd.gz to card root directory *");
	}
     else {
        CtlSetEnabled(ctlP,true);
        CtlShowControl(ctlP);
      // Initializing boot parameters
      kernelParametersField = (FieldType *)FrmGetObjectPtr(FrmGetActiveForm(), FrmGetObjectIndex(FrmGetActiveForm(), KernelParam));
      StrPrintF(buf,"mem=0x%08lX initrd=0x%08lX,%ld root=/dev/ram0 init=/linuxrc",
        PHYSICAL_RAM_SIZE,RAM_BASE_ADDRESS+0x400000,initrd_size);
      SetFieldTextFromStr(kernelParametersField, buf, true);
        lstP = FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, PortList));
	LstSetSelection(lstP,DEFAULT_PORT_INDEX);
	OnPortListChange();
	}
      
    }

/* --------------------------------------------------------------*/
int runLinuxKernel() {
  Err retcode;
  FormPtr frmP;
  ListPtr lstP;
  UInt32 creator=3007;
  UInt16 featureNum=35;
  UInt32 offset=0,bcount;
  MemPtr kernel;
  void *kernelPointer;
  MemHandle asmloaderHandle;
  void *asmloader;
  char *cline;
  UInt16 portid=0,i;
  UInt32 plist[]={serPortLocalHotSync,serPortCradlePort,serPortIrPort,
	    serPortConsolePort,serPortCradleRS232Port,serPortCradleUSBPort};
  FileRef fRef;
  Err err;

  //MemGluePtrNew() on palm tt2 allocate <750Kb,FtrPtrNew() allocate more

//  retcode = FtrPtrNew (creator,featureNum,krnl_size+initrd_size+0x20000,(void**)&kernel);
  retcode = FtrPtrNew (creator,featureNum,0x200000,(void**)&kernel);
  
  if (retcode != 0) {
    FrmCustomAlert(ErrorAlert,"Cannot allocate memory!",0,0);
    return -1;
  }

/********************************/
    frmP = FrmGetActiveForm();
    lstP = FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, PortList));
    if (lstP) {
	i=LstGetSelection(lstP);
	if ((i>0) && (i<8)) {
	    Log("Opening port...");
	    if (SrmOpen(plist[i-1], PORT_BAUD_RATE, &portid)!= errNone) {
	    FrmCustomAlert(ErrorAlert,"Cannot open port!",0,0);
	    return -1;
	    }
	SrmSend(portid, "GaruxNG: Port open under PalmOS.\n", 33, &err);
	}
    }
/********************************/

  // Fetch kernel command line
  cline=FldGetTextPtr(kernelParametersField);
  DmWrite(kernel, offset, cline, StrLen(cline)+1);
  DmWrite(kernel, offset+256-4, &krnl_size, 4);
  DmWrite(kernel, offset+256-8, &initrd_size, 4);
  offset += 256;

  
  kernelPointer = kernel;

  Log("Loading kernel...");
    retcode=VFSFileOpen(krnl_volRefNum,"/zImage",vfsModeRead,&fRef);
    if (retcode==errNone) {
      VFSFileSeek(fRef,vfsOriginBeginning,0);
      VFSFileReadData(fRef,krnl_size,kernel,offset,&bcount);
	VFSFileClose(fRef);
	if (bcount != krnl_size) {
    	    FrmCustomAlert(ErrorAlert,"Read failed [kernel]!",0,0);
	    }
	offset+=krnl_size;
	Log("Loading initrd...");
	retcode=VFSFileOpen(krnl_volRefNum,"/initrd.gz",vfsModeRead,&fRef);
	if (retcode==errNone) {
	    VFSFileSeek(fRef,vfsOriginBeginning,0);
	    VFSFileReadData(fRef,initrd_size,kernel,offset,&bcount);
	    if (bcount != initrd_size) {
    	       FrmCustomAlert(ErrorAlert,"Read failed [initrd]!",0,0);
	     }
	    VFSFileClose(fRef);
	}
    }

    
  if (retcode ==0) {
	Log("Booting Linux...");
	SysTaskDelay( 1*SysTicksPerSecond() );
	asmloaderHandle = DmGetResource('armc', 0);
	asmloader = MemHandleLock(asmloaderHandle);
	PceNativeCall(asmloader, kernel);
	MemHandleUnlock (asmloaderHandle);
    }

  // The following code shall not be executed
  FrmCustomAlert(ErrorAlert,"Boot failed!",0,0);
  FtrPtrFree(creator,featureNum);
  return 0;
}

/* --------------------------------------------------------------*/
void SetFieldTextFromStr( FieldPtr fldP, char *strP, Boolean redraw)
{
	MemHandle	oldTxtH;
	MemHandle	txtH;
	MemPtr		txtP;
	Boolean		fieldWasEditable;
	FieldAttrType	attrs;

	if (fldP==NULL) return;

	// get some space in which to stash the string.
	txtH	= MemHandleNew(  StrLen( strP ) + 1 );
	txtP	= MemHandleLock( txtH );
	// copy the string.
	StrCopy( (char*) txtP, strP );
	MemHandleUnlock( txtH );
	
	FldGetAttributes(fldP,&attrs);
	fieldWasEditable = attrs.editable;
	attrs.editable	 = true;
	FldSetAttributes(fldP,&attrs);

	oldTxtH	= FldGetTextHandle( fldP );
	FldSetTextHandle( fldP, txtH );
	
	attrs.editable	= fieldWasEditable;
	FldSetAttributes(fldP,&attrs);

	if ( oldTxtH != NULL ) MemHandleFree( oldTxtH );

	if (redraw) FldDrawField(fldP);
}
void OnPortListChange() {
ControlPtr ctlP;
ListPtr lstP;
FormPtr frmP=FrmGetActiveForm();

        frmP = FrmGetActiveForm();
	ctlP = FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, PortTrigger));
	lstP = FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, PortList));
	if (ctlP && lstP) {
	    CtlSetLabel (ctlP,LstGetSelectionText (lstP, LstGetSelection(lstP)));
	    }
}
